# tqdm_dask_distributed.py

"""
The below code is from Alex Papanicolaou (https://github.com/alexifm), posted in
this comment: https://github.com/tqdm/tqdm/issues/278#issuecomment-507006253
It inherits the MIT license of the tqdm project, copied in full below.

TODO: Remove this file once https://github.com/tqdm/tqdm/issues/1230 is merged

MIT License (MIT)
-----------------

Copyright (c) 2019 Alex Papanicolaou (alexifm)

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import tqdm
from tornado.ioloop import IOLoop
from distributed.utils import LoopRunner, is_kernel
from distributed.client import futures_of
from distributed.diagnostics.progressbar import ProgressBar

class TqdmProgressBar(ProgressBar):
    def __init__(
        self,
        keys,
        scheduler=None,
        interval="100ms",
        loop=None,
        complete=True,
        start=True,
        **tqdm_kwargs
    ):
        super(TqdmProgressBar, self).__init__(
            keys, scheduler, interval, complete)
        self.tqdm = tqdm.tqdm(keys, **tqdm_kwargs)
        self.loop = loop or IOLoop()

        if start:
            loop_runner = LoopRunner(self.loop)
            loop_runner.run_sync(self.listen)

    def _draw_bar(self, remaining, all, **kwargs):
        update_ct = (all - remaining) - self.tqdm.n
        self.tqdm.update(update_ct)

    def _draw_stop(self, **kwargs):
        self.tqdm.close()

class TqdmNotebookProgress(ProgressBar):
    def __init__(
        self,
        keys,
        scheduler=None,
        interval="100ms",
        loop=None,
        complete=True,
        start=True,
        **tqdm_kwargs
    ):
        super(TqdmNotebookProgress, self).__init__(
            keys, scheduler, interval, complete)
        self.tqdm = tqdm.tqdm_notebook(keys, **tqdm_kwargs)
        self.loop = loop or IOLoop()

        if start:
            loop_runner = LoopRunner(self.loop)
            loop_runner.run_sync(self.listen)

    def _draw_bar(self, remaining, all, **kwargs):
        update_ct = (all - remaining) - self.tqdm.n
        self.tqdm.update(update_ct)

    def _draw_stop(self, **kwargs):
        self.tqdm.close()

def tqdm_dask(futures, **kwargs):
    notebook = is_kernel()
    futures = futures_of(futures)
    if not isinstance(futures, (set, list)):
        futures = [futures]
    if notebook:
        return TqdmNotebookProgress(futures, **kwargs)
    else:
        TqdmProgressBar(futures, **kwargs)



# mc_enums.py
from __future__ import annotations

from enum import Enum

class SampleMethod(str, Enum):
    """
    Enum for the possible Monte Carlo sampling methods.
    """
    RANDOM          = 'random'
    SOBOL           = 'sobol'
    SOBOL_RANDOM    = 'sobol_random'
    HALTON          = 'halton'
    HALTON_RANDOM   = 'halton_random'
    LATIN_HYPERCUBE = 'latin_hypercube'


class SimFunctions(str, Enum):
    """
    Enum for the three required user functions.

    Notes
    -----
    The preprocess function must take in only a moncao.mc_case.Case object. It
    then must return a tuple of the input arguments for the run function.
    The run function will take in whatever inputs and return whatever outputs.
    It it recommended to package the outputs into a tuple.
    The postprocess function must take in as its first argument a
    moncao.mc_case.Case object, followed by the outputs from the run function.
    The simulation will attempt to unpack the run function outputs if they are
    stored in a tuple.

    See
    https://github.com/scottshambaugh/monaco/blob/main/template/template_functions.py
    for an example of this.
    """
    PREPROCESS  = 'preprocess'
    RUN         = 'run'
    POSTPROCESS = 'postprocess'


class StatBound(str, Enum):
    """
    Enum for possible statistical bounds. Note that not all of these may be
    valid for a given statistical function.
    """
    NEAREST        = 'nearest'
    BOTH           = 'both'
    ALL            = 'all'
    ONESIDED       = '1-sided'
    TWOSIDED       = '2-sided'
    ONESIDED_UPPER = '1-sided upper'
    ONESIDED_LOWER = '1-sided lower'


class VarStatType(str, Enum):
    """
    Enum for the variable statistics functions.
    """
    MAX         = 'max'
    MIN         = 'min'
    MEDIAN      = 'median'
    MEAN        = 'mean'
    GEOMEAN     = 'geomean'
    MODE        = 'mode'
    VARIANCE    = 'variance'
    SKEWNESS    = 'skewness'
    KURTOSIS    = 'kurtosis'
    MOMENT      = 'moment'
    PERCENTILE  = 'percentile'
    SIGMA       = 'sigma'
    GAUSSIANP   = 'gaussianp'
    ORDERSTATTI = 'orderstatti'
    ORDERSTATP  = 'orderstatp'


class VarStatSide(str, Enum):
    """
    Enum for the variable statistics 'side' (see documentation for each varstat
    function).
    """
    HIGH = 'high'
    LOW  = 'low'
    BOTH = 'both'
    ALL  = 'all'


class PlotOrientation(str, Enum):
    """
    Enum for the plotting functions orientation.
    """
    VERTICAL   = 'vertical'
    HORIZONTAL = 'horizontal'


class InVarSpace(str, Enum):
    """
    Enum for whether to plot invars in number or percentile space.
    """
    NUMS = 'nums'
    PCTS = 'pcts'


class Sensitivities(str, Enum):
    """
    Enum for whether to plot invars in number or percentile space.
    """
    INDICES = 'indices'
    RATIOS = 'ratios'



# flake8: noqa
from importlib import metadata
__version__ = metadata.version(__name__)

from monaco.mc_case import *
from monaco.mc_sim import *
from monaco.mc_var import *
from monaco.mc_varstat import *
from monaco.mc_val import *
from monaco.mc_enums import *
from monaco.gaussian_statistics import *
from monaco.order_statistics import *
from monaco.integration_statistics import *
from monaco.dvars_sensitivity import *
from monaco.helper_functions import *
from monaco.case_runners import *
from monaco.mc_sampling import *
from monaco.mc_plot import *
from monaco.mc_multi_plot import *
from monaco.tqdm_dask_distributed import *



# mc_sampling.py
from __future__ import annotations

import scipy.stats
import numpy as np
from functools import lru_cache
import warnings
import sys
from monaco.mc_enums import SampleMethod


def sampling(ndraws     : int,
             method     : SampleMethod = SampleMethod.SOBOL_RANDOM,
             ninvar     : int | None   = None,
             ninvar_max : int | None   = None,
             seed       : int          = np.random.get_state(legacy=False)['state']['key'][0],
             ) -> np.ndarray:
    """
    Draws random samples according to the specified method.

    Parameters
    ----------
    ndraws : int
        The number of samples to draw.
    method : monaco.mc_enums.SampleMethod
        The sample method to use.
    ninvar : int
        For all but the 'random' method, must define which number input
        variable is being sampled, ninvar >= 1. The 'sobol' and
        'sobol_random' methods must have ninvar <= 21201
    ninvar_max : int
        The total number of invars, ninvar_max >= ninvar. Used for caching.
    seed : int, default: np.random.get_state(legacy=False)['state']['key'][0]
        The random seed. Not used in 'sobol' or 'halton' methods.

    Returns
    -------
    pcts : numpy.ndarray
        The random samples. Each sample is 0 <= pct <= 1.
    """
    if ninvar_max is None:
        ninvar_max = ninvar

    if method == SampleMethod.RANDOM:
        pcts = scipy.stats.uniform.rvs(size=ndraws, random_state=seed)

    elif method in (SampleMethod.SOBOL, SampleMethod.SOBOL_RANDOM,
                    SampleMethod.HALTON, SampleMethod.HALTON_RANDOM, SampleMethod.LATIN_HYPERCUBE):
        if ninvar is None:
            raise ValueError(f'{ninvar=} must defined for the {method} method')
        elif (not 1 <= ninvar <= 21201) and method in (SampleMethod.SOBOL,
                                                       SampleMethod.SOBOL_RANDOM):
            raise ValueError(f'{ninvar=} must be between 1 and 21201 for the {method} method')

        scramble = False
        if method in (SampleMethod.SOBOL_RANDOM, SampleMethod.HALTON_RANDOM):
            scramble = True
        elif method in (SampleMethod.SOBOL, SampleMethod.HALTON):
            seed = 0  # These do not use randomness, so keep seed constant for caching

        all_pcts = cached_pcts(ndraws=ndraws, method=method, ninvar_max=ninvar_max,
                               scramble=scramble, seed=seed)
        pcts = all_pcts[:, ninvar-1]  # ninvar will always be >= 1

    else:
        raise ValueError("".join([f'{method=} must be one of the following: ' +
                                  f'{SampleMethod.RANDOM}, {SampleMethod.SOBOL}, ' +
                                  f'{SampleMethod.SOBOL_RANDOM}, {SampleMethod.HALTON}, ' +
                                  f'{SampleMethod.HALTON_RANDOM}, {SampleMethod.LATIN_HYPERCUBE}']))

    return pcts


@lru_cache(maxsize=1)
def cached_pcts(ndraws     : int,
                method     : str,
                ninvar_max : int,
                scramble   : bool,
                seed       : int,
                ) -> np.ndarray:
    """
    Wrapper function to cache the qmc draws so that we don't repeat calculation
    of lower numbered invars for the higher numbered invars.

    Parameters
    ----------
    ndraws : int
        The number of samples to draw.
    method : monaco.mc_enums.SampleMethod
        The sample method to use.
    ninvar_max : int
        The total number of invars.
    scramble : bool
        Whether to scramble the sobol or halton points. Should only be True if
        method is in {'sobol_random', 'halton_random'}
    seed : int
        The random seed. Not used in 'sobol' or 'halton' methods.

    Returns
    -------
    all_pcts : numpy.ndarray
        The random samples. Each sample is 0 <= pct <= 1.
    """
    if method in (SampleMethod.SOBOL, SampleMethod.SOBOL_RANDOM):
        sampler = scipy.stats.qmc.Sobol(d=ninvar_max, scramble=scramble, seed=seed)
    elif method in (SampleMethod.HALTON, SampleMethod.HALTON_RANDOM):
        sampler = scipy.stats.qmc.Halton(d=ninvar_max, scramble=scramble, seed=seed)
    elif method == SampleMethod.LATIN_HYPERCUBE:
        sampler = scipy.stats.qmc.LatinHypercube(d=ninvar_max, seed=seed)

    if not sys.warnoptions:
        with warnings.catch_warnings():
            # Suppress the power of 2 warning for sobol / halton sequences
            warnings.simplefilter("ignore", category=UserWarning)
            points = sampler.random(n=ndraws)
    else:
        points = sampler.random(n=ndraws)

    all_pcts = np.array(points)
    return all_pcts



# mc_case.py
from __future__ import annotations

from monaco.mc_var import OutVar, InVar
from monaco.mc_val import OutVal, InVal
from datetime import datetime, timedelta
from pathlib import Path
from typing import Any
import numpy as np

class Case():
    """
    Object to hold all the data for a single Monte Carlo case.

    Parameters
    ----------
    ncase : int
        The number of this case.
    ismedian : bool
        Whether this represents the median case.
    invars : dict[str, monaco.mc_var.InVar]
        A dict pointing to all of the input variables.
    constvals : dict[str, Any], default: None
        A dict of any constant values common to all cases.
    keepsiminput : bool, default: True
        Whether to keep the siminput after running.
    keepsimrawoutput : bool, default: True
        Whether to keep the simrawoutput after postprocessing.
    seed : int, default: np.random.get_state(legacy=False)['state']['key'][0]
        The random seed to pass to the run function for this case. Not used in
        as part of any Monte Carlo sampling.

    Attributes
    ----------
    starttime : datetime.datetime
        Timestamp for when this case started running.
    endtime : datetime.datetime
        Timestamp for when this case stopped running.
    runtime : datetime.timedelta
        Total run duration for this case.
    filepath : pathlib.Path
        The filepath for the case data, if saved to disk.
    runsimid : int
        The id for the particular sim run this case was run for.
    haspreprocessed : bool
        Whether this case has been preprocessed.
    hasrun : bool
        Whether this case has run the run function.
    haspostprocessed : bool
        Whether this case has been postprocessed.
    invals : dict[str, monaco.mc_val.InVal]
        The input values for this partitcular case.
    outvals : dict[str, monaco.mc_val.OutVal]
        The output values for this partitcular case.
    siminput : tuple[Any]
        The preprocessed inputs provided to the run function for this case.
    simrawoutput : tuple[Any]
        The non-postprocessed outputs from the run function for this case.
    """
    def __init__(self,
                 ncase            : int,
                 ismedian         : bool,
                 invars           : dict[str, InVar],
                 constvals        : dict[str, Any] | None = None,
                 keepsiminput     : bool = True,
                 keepsimrawoutput : bool = True,
                 seed             : int = np.random.get_state(legacy=False)['state']['key'][0],
                 ):

        self.ncase = ncase
        self.ismedian = ismedian
        self.invars = invars
        if constvals is None:
            constvals = dict()
        self.constvals = constvals
        self.outvars : dict[str, OutVar] = dict()
        self.keepsiminput = keepsiminput
        self.keepsimrawoutput = keepsimrawoutput
        self.seed = seed

        self.starttime : datetime | None = None
        self.endtime   : datetime | None = None
        self.runtime   : timedelta | None = None

        self.filepath : Path | None = None
        self.runsimid : int | None = None
        self.haspreprocessed  : bool = False
        self.hasrun           : bool = False
        self.haspostprocessed : bool = False

        self.invals  : dict[str, InVal]  = self.getInVals()
        self.outvals : dict[str, OutVal] = dict()

        self.siminput     : tuple[Any] | None = None
        self.simrawoutput : tuple[Any] | None = None


    def __repr__(self):
        return (f"{self.__class__.__name__}(ncase={self.ncase},\n  invals={self.invals} " +
                f"\n  outvals={self.outvals})")


    def getInVals(self) -> dict[str, InVal]:
        """
        From all the InVar's, extract the vals for this case.

        Returns
        -------
        vals: dict[str, monaco.mc_val.InVal]
            The InVal's for this case.
        """
        vals = dict()
        for var in self.invars.values():
            val = var.getVal(self.ncase)
            vals[val.name] = val
        return vals


    def getOutVals(self) -> dict[str, OutVal]:
        """
        From all the OutVar's, extract the vals for this case.

        Returns
        -------
        vals: dict[str, monaco.mc_val.OutVal]
            The OutVal's for this case.
        """
        vals = dict()
        for var in self.outvars.values():
            val = var.getVal(self.ncase)
            vals[val.name] = val
        return vals


    def addOutVal(self,
                  name   : str,
                  val    : Any,
                  split  : bool = True,
                  valmap : dict[Any, float] | None = None
                  ) -> None:
        """
        Generate an OutVal and add it to the dict of outvals.

        Parameters
        ----------
        name : str
            The name of the output value.
        val : Any
            The output value.
        split : bool
            Whether to split the value into its components.
        valmap : dict[Any, float], default: None
            A valmap dict mapping nonnumeric values to numbers.
        """
        if name in self.outvals.keys():
            raise ValueError(f"'{name}' is already an OutVal")

        self.outvals[name] = OutVal(name=name, ncase=self.ncase, val=val,
                                    valmap=valmap, ismedian=self.ismedian)
        if split:
            self.outvals.update(self.outvals[name].split())



# integration_statistics.py
from __future__ import annotations

import numpy as np
from scipy.optimize import root_scalar
from typing import Optional
from monaco.gaussian_statistics import pct2sig
from monaco.mc_enums import SampleMethod

def integration_error(nums         : np.ndarray,
                      dimension    : int,
                      volume       : float        = 1,
                      conf         : float        = 0.95,
                      samplemethod : SampleMethod = SampleMethod.RANDOM,
                      runningerror : bool         = False,
                      ) -> float | np.ndarray:
    """
    Returns the bounding integration error for an input array of numbers. This
    error can be a float point estimate if runningerror == False, or a numpy
    array showing running error over the samples, to demonstrate convergence.
    If volume == 1, the error returned is a percent error. Otherwise, the error
    is absolute over the integration volume.

    Parameters
    ----------
    nums : list[float]
        A list of the integration point estimates across the volume.
    dimension : int
        The number of dimensions in the integration volume, dimension > 0.
    volume : float
        The integration volume, > 0. If volume == 1 (default), then the error
        returned is a percentage of the true integration volume.
    conf : float
        Confidence level of the calculated error. This must be 0 < conf < 1,
        and should be 0.5 < conf < 1.
    samplemethod : monaco.mc_enums.SampleMethod
        Monte Carlo sample method. Either 'random' (default and bounding), or
        'sobol'. If using a different sample method, use 'random' here.
    runningerror : bool
        If False, returns a point estimate. If True, returns an array
        containing the running error over all of the integration estimates.

    Returns
    -------
    error : float | np.ndarray
        Either a point estimate of the error if runningerror == False, or an
        array of the running error if runningerror is True.
    """
    integration_args_check(error=None, dimension=dimension, volume=volume,
                           stdev=None, conf=conf, samplemethod=samplemethod)

    n = len(nums)
    if n == 1:
        error1sig = np.array(volume)

    elif not runningerror:
        stdev = np.std(nums, ddof=1)
        error1sig_random = volume*np.sqrt((2**(-1*dimension) - 3**(-1*dimension))/n)
        if samplemethod == SampleMethod.RANDOM:
            error1sig = error1sig_random
        elif samplemethod == SampleMethod.SOBOL:
            error1sig_sobol = volume*stdev*np.log(n)**dimension/n
            error1sig = np.minimum(error1sig_random, error1sig_sobol)

    else:
        # Use Welford's algorithm to calculate the running variance
        M = np.zeros(n)  # Running mean
        S = np.zeros(n)  # Sum of variances
        M[0] = nums[0]
        for i in range(1, n):
            M[i] = M[i-1] + (nums[i]-M[i-1])/(i+1)
            S[i] = S[i-1] + (nums[i]-M[i-1])*(nums[i]-M[i])
        variances = np.zeros(n)
        variances[1:] = S[1:]/np.arange(1, n)
        stdevs = np.sqrt(variances)

        error1sig_random = volume*np.sqrt((2**(-1*dimension) - 3**(-1*dimension))/np.arange(1, n+1))
        if samplemethod == SampleMethod.RANDOM:
            error1sig = error1sig_random
        elif samplemethod == SampleMethod.SOBOL:
            error1sig_sobol = volume*stdevs*np.log(np.arange(1, n+1))**dimension/np.arange(1, n+1)
            error1sig = np.minimum(error1sig_random, error1sig_sobol)

        # Leading zeros will throw off plots, fill with reasonable dummy data
        error1sig[error1sig == 0] = np.max(error1sig)

    error : float | np.ndarray = error1sig*pct2sig(conf)
    return error


def integration_n_from_err(error        : float,
                           dimension    : int,
                           volume       : float,
                           stdev        : float,
                           conf         : float        = 0.95,
                           samplemethod : SampleMethod = SampleMethod.RANDOM,
                           ) -> int:
    """
    Returns the bounding integration error for an input array of numbers. This
    error can be a float point estimate if runningerror == False, or a numpy
    array showing running error over the samples, to demonstrate convergence.
    If volume == 1, the error returned is a percent error. Otherwise, the error
    is absolute over the integration volume.
    We generally do not know a-priori what the standard deviation will be, so
    best practice is to set to the max range of values on the interval, and
    then calculate a better stdev on a lower number of cases, which can then
    be subsituted in here to bootleg a more efficient computation.
    For sobol sampling, remember to round n to the next power of 2 for balance.
    monaco.helper_functions.next_power_of_2(n) can help with this.

    Parameters
    ----------
    error : float
        The target error.
    dimension : int
        The number of dimensions in the integration volume, dimension > 0.
    volume : float
        The integration volume, > 0. If volume == 1 (default), then the target
        error is a percentage of the true integration volume.
    stdev : float
        The standard deviation of the integration estimates, stdev > 0. We
        generally do not know this a-priori, so use
        monaco.integration_statistics.max_stdev to calculate this in that
        instance. Or, do a limited number of cases to estimate this before
        performing the full run.
    conf : float
        Confidence level of the calculated error. This must be 0 < conf < 1,
        and should be 0.5 < conf < 1.
    samplemethod : monaco.mc_enums.SampleMethod
        Monte Carlo sample method. Either 'random' (default and bounding), or
        'sobol'. If using a different sample method, use 'random' here.

    Returns
    -------
    n : int
        The number of sample points required to meet the target integration
        error.
    """
    integration_args_check(error=error, dimension=dimension, volume=volume,
                           stdev=stdev, conf=conf, samplemethod=samplemethod)

    n_random = (volume*pct2sig(conf)*stdev/error)**2
    if samplemethod == SampleMethod.RANDOM:
        n = n_random
    elif samplemethod == SampleMethod.SOBOL:
        def f(n : float) -> float:
            return volume*stdev*pct2sig(conf)*np.log(n)**dimension/n - error

        try:
            rootResults = root_scalar(f, method='brentq', bracket=[2**8, 2**31-1],
                                         xtol=0.1, maxiter=int(1e3))
            n_sobol = rootResults.root
            n = np.min([n_random, n_sobol])
        except Exception:
            # For higher than 3 dimensions, reaching n may be difficult, and
            # will be much larger than n_random anyways
            # warn(f'Cannot reach error tolerance of ±{error}. ' +
            #      f'Falling back to samplemethod={SampleMethod.RANDOM}')
            n = n_random

    n = int(np.ceil(n))
    return n


def integration_args_check(error        : Optional[float],
                           dimension    : int,
                           volume       : float,
                           stdev        : Optional[float],
                           conf         : float,
                           samplemethod : SampleMethod,
                           ) -> None:
    """
    Raises a ValueError if any of the inputs for the integration functions are
    outside allowable bounds.

    Parameters
    ----------
    error : None | float
        error > 0.
    dimension : int
        dimension > 0.
    volume : float
        volume > 0.
    stdev : None | float
        stdev > 0.
    conf : float
         0 < conf < 1
    samplemethod : monaco.mc_enums.SampleMethod
        Either 'random' or 'sobol'.
    """
    if (error is not None) and (error < 0):
        raise ValueError(f'{error=} must be positive')
    if dimension < 1:
        raise ValueError(f'{dimension=} must be a positive integer')
    if volume <= 0:
        raise ValueError(f'{volume=} must be positive')
    if (stdev is not None) and (stdev < 0):
        raise ValueError(f'{stdev=} must be positive')
    if not 0 < conf < 1:
        raise ValueError(f'{conf=} must be between 0 and 1')
    if samplemethod not in (SampleMethod.RANDOM, SampleMethod.SOBOL):
        raise ValueError(f'{samplemethod=} must be either ' +
                         f'{SampleMethod.RANDOM} or {SampleMethod.SOBOL}')


def max_variance(low  : float,
                 high : float,
                 ) -> float:
    """
    Calculates the maximum possible variance of a list of points that span the
    range [low, high].
    max_variance(0, 1) = 0.25

    Parameters
    ----------
    low : float
        The low end of the range.
    high : float
        The high end of the range.

    Returns
    -------
    maxvar: float
        The maximum possible variance of a list of points in the input range.
    """
    maxvar = (high-low)**2 / 4
    return maxvar


def max_stdev(low  : float,
              high : float,
              ) -> float:
    """
    Calculates the maximum possible variance of a list of points that span the
    range [low, high].
    max_stdev(0, 1) = 0.5

    Parameters
    ----------
    low : float
        The low end of the range.
    high : float
        The high end of the range.

    Returns
    -------
    maxstd: float
        The maximum possible standard deviation of a list of points in the
        input range.
    """
    maxstd = np.sqrt(max_variance(high=high, low=low))  # maxstd = (high-low)/2
    return maxstd



# mc_plot.py
from __future__ import annotations

# Somewhat hacky type checking to avoid circular imports:
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from monaco.mc_var import InVar, OutVar
import monaco.mc_var

import numpy as np
import matplotlib.pyplot as plt
from scipy.interpolate import griddata
from scipy.stats import rv_continuous, rv_discrete, chi2, mode
from copy import copy, deepcopy
from typing import Optional, Iterable
from matplotlib.figure import Figure
from matplotlib.axes import Axes
from matplotlib.patches import Ellipse
from mpl_toolkits.mplot3d import Axes3D  # noqa: F401
from monaco.helper_functions import get_list, slice_by_index, length, empty_list
from monaco.gaussian_statistics import conf_ellipsoid_sig2pct
from monaco.integration_statistics import integration_error
from monaco.mc_enums import SampleMethod, PlotOrientation, InVarSpace, Sensitivities


# If cases or highlight_cases are None, will plot all. Set to [] to plot none.
def plot(varx   : InVar | OutVar,
         vary   : InVar | OutVar = None,
         varz   : InVar | OutVar = None,
         cases           : None | int | Iterable[int] = None,
         highlight_cases : None | int | Iterable[int] = empty_list(),
         rug_plot    : bool           = False,
         cov_plot    : bool           = False,
         cov_p       : None | float | Iterable[float] = None,
         invar_space : InVarSpace | Iterable[InVarSpace] = InVarSpace.NUMS,
         ax          : Optional[Axes] = None,
         title       : str            = '',
         plotkwargs  : dict           = dict(),
         ) -> tuple[Figure, Axes]:
    """
    Umbrella function to make single plots of a single Monte Carlo variable or
    pairs or triplets of variables.

    Parameters
    ----------
    varx : monaco.mc_var.InVar | monaco.mc_var.OutVar
        The x variable to plot.
    vary : monaco.mc_var.InVar | monaco.mc_var.OutVar, default: None
        The y variable to plot.
    varz : monaco.mc_var.InVar | monaco.mc_var.OutVar, default: None
        The z variable to plot.
    cases : None | int | Iterable[int], default: None
        The cases to plot. If None, then all cases are plotted.
    highlight_cases : None | int | Iterable[int], default: []
        The cases to highlight. If [], then no cases are highlighted.
    rug_plot : bool, default: True
        Whether to plot rug marks.
    cov_plot : bool, default: False
        Whether to plot a covariance ellipse at a certain gaussian percentile
        level.
    cov_p : None | float | Iterable[float], default: None
        The gaussian percentiles for the covariance plot.
    invar_space : monaco.InVarSpace | Iterable[InVarSpace], default: 'nums'
        The space to plot invars in, either 'nums' or 'pcts'. If an iterable,
        specifies this individually for each of varx, vary, and varz.
    ax : matplotlib.axes.Axes, default: None
        The axes handle to plot in. If None, a new figure is created.
    title : str, default: ''
        The figure title.

    Returns
    -------
    (fig, ax) : (matplotlib.figure.Figure, matplotlib.axes.Axes)
        fig is the figure handle for the plot.
        ax is the axes handle for the plot.
    """
    # Split single vars that are 2D or 3D vectors
    if vary is None and varz is None:
        if varx.maxdim not in (0, 1, 2):
            raise ValueError(f'Invalid variable dimension: {varx.name}: {varx.maxdim}')

        elif varx.maxdim == 2 and isinstance(varx, monaco.mc_var.OutVar):
            varx_split = varx.split()  # split only defined for OutVar
            origname = varx.name
            varx = varx_split[origname + ' [0]']
            vary = varx_split[origname + ' [1]']
            if len(varx_split) == 3:
                varz = varx_split[origname + ' [2]']
            elif len(varx_split) > 3:
                raise ValueError('Can only split a single variable into 3 vectors:' +
                                 f'{origname}: {len(varx_split)}')

    # Split one of two vars that is a 2D vector
    elif vary is not None and varz is None:
        if varx.maxdim == 2 and vary.maxdim == 0 and isinstance(varx, monaco.mc_var.OutVar):
            varx_split = varx.split()  # split only defined for OutVar
            origname = varx.name
            varz = vary
            varx = varx_split[origname + ' [0]']
            vary = varx_split[origname + ' [1]']
            if len(varx_split) > 2:
                raise ValueError('Can only split one of two variables into 2 vectors:' +
                                 f'{origname}: {len(varx_split)}')

        elif varx.maxdim == 0 and vary.maxdim == 2 and isinstance(vary, monaco.mc_var.OutVar):
            vary_split = vary.split()  # split only defined for OutVar
            origname = vary.name
            vary = vary_split[origname + ' [0]']
            varz = vary_split[origname + ' [1]']
            if len(vary_split) > 2:
                raise ValueError('Can only split one of two variables into 2 vectors:' +
                                 f'{origname}: {len(vary_split)}')

    # Single Variable Plots
    if vary is None and varz is None:
        if varx.maxdim == 0:
            fig, ax = plot_hist(var=varx, cases=cases, highlight_cases=highlight_cases,
                                rug_plot=rug_plot, invar_space=invar_space, ax=ax, title=title,
                                plotkwargs=plotkwargs)
        else:
            var1 = copy(varx)
            var0 = get_var_steps(varx)
            fig, ax = plot_2d_line(varx=var0, vary=var1,
                                   highlight_cases=highlight_cases,
                                   invar_space=invar_space,
                                   ax=ax, title=title, plotkwargs=plotkwargs)

    # Two Variable Plots
    elif vary is not None and varz is None:
        if varx.maxdim == 0 and vary.maxdim == 0:
            fig, ax = plot_2d_scatter(varx=varx, vary=vary,
                                      cases=cases, highlight_cases=highlight_cases,
                                      rug_plot=rug_plot, cov_plot=cov_plot, cov_p=cov_p,
                                      invar_space=invar_space,
                                      ax=ax, title=title, plotkwargs=plotkwargs)

        elif varx.maxdim == 1 and vary.maxdim == 1:
            fig, ax = plot_2d_line(varx=varx, vary=vary,
                                   cases=cases, highlight_cases=highlight_cases,
                                   invar_space=invar_space,
                                   ax=ax, title=title, plotkwargs=plotkwargs)

        elif varx.maxdim in (0, 1) and vary.maxdim in (0, 1):
            var2 = copy(vary)
            if varx.maxdim == 1:
                var1 = varx
                var0 = get_var_steps(varx)
            elif vary.maxdim == 1:
                var1 = get_var_steps(vary)
                var0 = varx
            fig, ax = plot_2p5d_line(varx=var0, vary=var1, varz=var2,
                                     cases=cases, highlight_cases=highlight_cases,
                                     invar_space=invar_space,
                                     ax=ax, title=title, plotkwargs=plotkwargs)

        else:
            raise ValueError( 'Invalid variable dimension: ' +
                             f'{varx.name}: {varx.maxdim}, ' +
                             f'{vary.name}: {vary.maxdim}')

    # Three Variable Plots
    else:
        if varx.maxdim == 0 and vary.maxdim == 0 and varz.maxdim == 0:
            fig, ax = plot_3d_scatter(varx=varx, vary=vary, varz=varz,
                                      cases=cases, highlight_cases=highlight_cases,
                                      invar_space=invar_space,
                                      ax=ax, title=title, plotkwargs=plotkwargs)

        elif varx.maxdim == 1 and vary.maxdim == 1 and varz.maxdim == 1:
            fig, ax = plot_3d_line(varx=varx, vary=vary, varz=varz,
                                   cases=cases, highlight_cases=highlight_cases,
                                   invar_space=invar_space,
                                   ax=ax, title=title, plotkwargs=plotkwargs)

        elif varx.maxdim in (0, 1) and vary.maxdim in (0, 1) and varz.maxdim in (0, 1):
            fig, ax = plot_2p5d_line(varx=varx, vary=vary, varz=varz,
                                     cases=cases, highlight_cases=highlight_cases,
                                     invar_space=invar_space,
                                     ax=ax, title=title, plotkwargs=plotkwargs)

        else:
            raise ValueError( 'Invalid variable dimension: ' +
                             f'{varx.name}: {varx.maxdim}, ' +
                             f'{vary.name}: {vary.maxdim}, ' +
                             f'{varz.name}: {varz.maxdim}')

    return fig, ax



def plot_hist(var         : InVar | OutVar,
              cases           : None | int | Iterable[int] = None,
              highlight_cases : None | int | Iterable[int] = empty_list(),
              cumulative  : bool            = False,
              orientation : PlotOrientation = PlotOrientation.VERTICAL,
              rug_plot    : bool            = True,
              invar_space : InVarSpace | Iterable[InVarSpace] = InVarSpace.NUMS,
              ax          : Optional[Axes]  = None,
              title       : str             = '',
              plotkwargs  : dict            = dict(),
              ) -> tuple[Figure, Axes]:
    """
    Plot a histogram of a single variable.

    Parameters
    ----------
    var : monaco.mc_var.InVar | monaco.mc_var.OutVar
        The variable to plot.
    cases : None | int | Iterable[int], default: None
        The cases to plot. If None, then all cases are plotted.
    highlight_cases : None | int | Iterable[int], default: []
        The cases to highlight. If [], then no cases are highlighted.
    cumulative : bool, default: False
        Whether to plot the histograms as cumulative distribution functions.
    orientation : monaco.mc_enums.PlotOrientation, default: 'vertical'
        The orientation of the histogram. Either 'vertical' or 'horizontal'.
    rug_plot : bool, default: True
        Whether to plot rug marks.
    invar_space : monaco.InVarSpace | Iterable[InVarSpace], default: 'nums'
        The space to plot invars in, either 'nums' or 'pcts'. If an iterable,
        specifies this individually for each of varx, vary, and varz.
    ax : matplotlib.axes.Axes, default: None
        The axes handle to plot in. If None, a new figure is created.
    title : str, default: ''
        The figure title.

    Returns
    -------
    (fig, ax) : (matplotlib.figure.Figure, matplotlib.axes.Axes)
        fig is the figure handle for the plot.
        ax is the axes handle for the plot.
    """
    fig, ax = manage_axis(ax, is3d=False)
    invar_space = manage_invar_space(invar_space=invar_space, nvars=1)

    bins : str | np.ndarray = 'auto'
    if 'bins' in plotkwargs:
        bins = plotkwargs['bins']
        del plotkwargs['bins']

    # Histogram generation
    cases_list = get_cases(var.ncases, cases)
    highlight_cases_list = get_cases(var.ncases, highlight_cases)
    points = get_plot_points(var, invar_space[0])
    nums = slice_by_index(points, cases_list)
    counts, bins = np.histogram(nums, bins=bins)
    binwidth = mode(np.diff(bins), keepdims=False)[0]
    bins = np.append(bins - binwidth/2, bins[-1] + binwidth/2)
    counts, bins = np.histogram(nums, bins=bins)

    if isinstance(var, monaco.mc_var.InVar):
        # Loaded from file
        if var.dist is None:
            plt.hist(bins[:-1], bins=bins, weights=counts/sum(counts), density=True,
                     cumulative=cumulative, orientation=orientation, histtype='bar',
                     facecolor='k', alpha=0.5, **plotkwargs)

        # Continuous distribution
        if isinstance(var.dist, rv_continuous):
            plt.hist(bins[:-1], bins=bins, weights=counts/sum(counts), density=True,
                     cumulative=cumulative, orientation=orientation, histtype='bar',
                     facecolor='k', alpha=0.5, **plotkwargs)
            lim = get_hist_lim(ax, orientation)
            x = np.arange(lim[0], lim[1], (lim[1] - lim[0])/100)
            dist = var.dist(**var.distkwargs)
            if cumulative:
                ydata = dist.cdf(x)
            else:
                ydata = dist.pdf(x)
            if orientation == PlotOrientation.VERTICAL:
                plt.plot(x, ydata, color='k', alpha=0.9)
            elif orientation == PlotOrientation.HORIZONTAL:
                plt.plot(ydata, x, color='k', alpha=0.9)

        # Discrete distribution
        elif isinstance(var.dist, rv_discrete):
            plt.hist(bins[:-1], bins=bins, weights=counts/sum(counts), density=False,
                     orientation=orientation, cumulative=cumulative, histtype='bar',
                     facecolor='k', alpha=0.5, **plotkwargs)
            lim = get_hist_lim(ax, orientation)
            x = np.arange(int(np.floor(lim[0])), int(np.ceil(lim[1])))
            dist = var.dist(**var.distkwargs)
            xdata = x
            if cumulative:
                ydata = dist.cdf(x)
                if orientation == PlotOrientation.VERTICAL:
                    plt.step(xdata, ydata, color='k', alpha=0.9, where='post')
                elif orientation == PlotOrientation.HORIZONTAL:
                    plt.step(ydata, xdata, color='k', alpha=0.9, where='post')
            else:
                ydata = dist.pmf(x)
                if orientation == PlotOrientation.VERTICAL:
                    plt.vlines(xdata, 0, ydata, color='k', alpha=0.9)
                elif orientation == PlotOrientation.HORIZONTAL:
                    plt.hlines(xdata, 0, ydata, color='k', alpha=0.9)

    elif isinstance(var, monaco.mc_var.OutVar):
        plt.hist(bins[:-1], bins=bins, weights=counts/sum(counts), density=False,
                 orientation=orientation, cumulative=cumulative, histtype='bar',
                 facecolor='k', alpha=0.5, **plotkwargs)

    if cumulative:
        ylabeltext = 'Cumulative Probability'
    else:
        ylabeltext = 'Probability Density'

    if rug_plot:
        plot_rug_marks(ax, orientation=orientation, nums=np.array(points))

    xlim = ax.get_xlim()
    ylim = ax.get_ylim()

    # Highlight cases and VarStats
    if orientation == PlotOrientation.VERTICAL:
        for i in highlight_cases_list:
            plt.plot([points[i], points[i]],
                     [ylim[0], ylim[0] + (ylim[1] - ylim[0])*0.20],
                     linestyle='-', linewidth=1, color='C1', alpha=1,
                     **plotkwargs)
        for i, varstat in enumerate(var.varstats):
            nums = get_list(varstat.nums)
            if length(nums) == 1:  # Single Statistic
                plt.plot([nums[0], nums[0]], ylim,
                         linestyle='-', color='C0', alpha=1, **plotkwargs)
            elif length(nums) == 3:  # Sided Order Statistics
                plt.plot([nums[1], nums[1]], ylim,
                         linestyle='-', color='C0', alpha=1, **plotkwargs)
            if varstat.confidence_interval_high_nums is not None:
                ax.fill_between([varstat.confidence_interval_low_nums,
                                 varstat.confidence_interval_high_nums],
                                [ylim[0], ylim[0]],
                                [ylim[1], ylim[1]],
                                color='C0', alpha=0.3)
            if length(nums) in (2, 3):  # Sided Order Statistics
                ax.fill_between([nums[0], nums[-1]],
                                [ylim[0], ylim[0]],
                                [ylim[1], ylim[1]],
                                color='C0', alpha=0.3)
        plt.xlabel(var.name)
        plt.ylabel(ylabeltext)
        apply_category_labels(ax, varx=var)

    elif orientation == PlotOrientation.HORIZONTAL:
        for i in highlight_cases_list:
            plt.plot([xlim[0], xlim[0] + (xlim[1] - xlim[0])*0.20],
                     [points[i], points[i]],
                     linestyle='-', linewidth=1, color='C1', alpha=1, **plotkwargs)
        for varstat in var.varstats:
            nums = get_list(varstat.nums)
            if length(nums) == 1:  # Single Statistic
                plt.plot(xlim, [nums[0], nums[0]],
                         linestyle='-', color='C0', alpha=1, **plotkwargs)
            elif length(nums) == 3:  # Sided Order Statistics
                plt.plot(xlim, [nums[1], nums[1]],
                         linestyle='-', color='C0', alpha=1, **plotkwargs)
            if varstat.confidence_interval_high_nums is not None:
                ax.fill_between(xlim,
                                [varstat.confidence_interval_low_nums,
                                 varstat.confidence_interval_high_nums],
                                color='C0', alpha=0.3)
            if length(nums) in (2, 3):  # Sided Order Statistics
                ax.fill_between(xlim, nums[0], nums[-1], color='C0', alpha=0.3)
        plt.ylabel(var.name)
        plt.xlabel(ylabeltext)
        apply_category_labels(ax, vary=var)

    plt.title(title)

    return fig, ax



def plot_cdf(var       : InVar | OutVar,
             cases           : None | int | Iterable[int] = None,
             highlight_cases : None | int | Iterable[int] = empty_list(),
             orientation : PlotOrientation = PlotOrientation.VERTICAL,
             rug_plot    : bool            = True,
             invar_space : InVarSpace | Iterable[InVarSpace] = InVarSpace.NUMS,
             ax          : Optional[Axes]  = None,
             title       : str             = '',
             plotkwargs  : dict            = dict(),
             ) -> tuple[Figure, Axes]:
    """
    Plot a cumulative distribution of a single variable.

    Parameters
    ----------
    var : monaco.mc_var.InVar | monaco.mc_var.OutVar
        The variable to plot.
    cases : None | int | Iterable[int], default: None
        The cases to plot. If None, then all cases are plotted.
    highlight_cases : None | int | Iterable[int], default: []
        The cases to highlight. If [], then no cases are highlighted.
    orientation : monaco.mc_enums.PlotOrientation, default: 'vertical'
        The orientation of the histogram. Either 'vertical' or 'horizontal'.
    rug_plot : bool, default: True
        Whether to plot rug marks.
    invar_space : monaco.InVarSpace | Iterable[InVarSpace], default: 'nums'
        The space to plot invars in, either 'nums' or 'pcts'. If an iterable,
        specifies this individually for each of varx, vary, and varz.
    ax : matplotlib.axes.Axes, default: None
        The axes handle to plot in. If None, a new figure is created.
    title : str, default: ''
        The figure title.

    Returns
    -------
    (fig, ax) : (matplotlib.figure.Figure, matplotlib.axes.Axes)
        fig is the figure handle for the plot.
        ax is the axes handle for the plot.
    """
    return plot_hist(var=var, cases=cases, highlight_cases=highlight_cases, cumulative=True,
                     orientation=orientation, rug_plot=rug_plot, invar_space=invar_space,
                     ax=ax, title=title, plotkwargs=plotkwargs)



def plot_2d_scatter(varx   : InVar | OutVar,
                    vary   : InVar | OutVar,
                    varz   : InVar | OutVar = None,
                    cases           : None | int | Iterable[int] = None,
                    highlight_cases : None | int | Iterable[int] = empty_list(),
                    rug_plot : bool           = False,
                    cov_plot : bool           = False,
                    cov_p    : None | float | Iterable[float] = None,
                    invar_space : InVarSpace | Iterable[InVarSpace] = InVarSpace.NUMS,
                    ax       : Optional[Axes] = None,
                    title    : str            = '',
                    plotkwargs : dict         = dict(),
                    ) -> tuple[Figure, Axes]:
    """
    Plot a scatter plot of two variables.

    Parameters
    ----------
    varx : monaco.mc_var.InVar | monaco.mc_var.OutVar
        The x variable to plot.
    vary : monaco.mc_var.InVar | monaco.mc_var.OutVar
        The y variable to plot.
    varz : monaco.mc_var.InVar | monaco.mc_var.OutVar, default: None
        If not None, then sets the values for the underlying contour plot.
    cases : None | int | Iterable[int], default: None
        The cases to plot. If None, then all cases are highlighted.
    highlight_cases : None | int | Iterable[int], default: []
        The cases to highlight. If [], then no cases are highlighted.
    rug_plot : bool, default: True
        Whether to plot rug marks.
    cov_plot : bool, default: False
        Whether to plot a covariance ellipse at a certain gaussian percentile
        level.
    cov_p : None | float | Iterable[float], default: None
        The gaussian percentiles for the covariance plot.
    invar_space : monaco.InVarSpace | Iterable[InVarSpace], default: 'nums'
        The space to plot invars in, either 'nums' or 'pcts'. If an iterable,
        specifies this individually for each of varx, vary, and varz.
    ax : matplotlib.axes.Axes, default: None
        The axes handle to plot in. If None, a new figure is created.
    title : str, default: ''
        The figure title.

    Returns
    -------
    (fig, ax) : (matplotlib.figure.Figure, matplotlib.axes.Axes)
        fig is the figure handle for the plot.
        ax is the axes handle for the plot.
    """
    nvars = 2
    if varz is not None:
        nvars = 3

    fig, ax = manage_axis(ax, is3d=False)
    invar_space = manage_invar_space(invar_space=invar_space, nvars=nvars)

    cases_list = get_cases(varx.ncases, cases)
    highlight_cases_list = get_cases(varx.ncases, highlight_cases)
    reg_cases = set(cases_list) - set(highlight_cases_list)
    varx_points = get_plot_points(varx, invar_space[0])
    vary_points = get_plot_points(vary, invar_space[1])

    if varz is not None:
        ngrid = 200
        cmap = 'viridis'
        levels = 21

        xi = np.linspace(min(varx_points), max(varx_points), ngrid)
        yi = np.linspace(min(vary_points), max(vary_points), ngrid)
        varz_points = get_plot_points(varz, invar_space[2])
        zi = griddata((varx_points, vary_points), varz_points,
                      (xi[None, :], yi[:, None]), method='linear')
        contour = ax.contourf(xi, yi, zi, levels=levels, cmap=cmap)
        fig.colorbar(contour, ax=ax, label=varz.name)

    if reg_cases:
        plt.scatter(slice_by_index(varx_points, reg_cases),
                    slice_by_index(vary_points, reg_cases),
                    edgecolors=None, c='k', alpha=0.4, **plotkwargs)
    if highlight_cases_list:
        plt.scatter(slice_by_index(varx_points, highlight_cases_list),
                    slice_by_index(vary_points, highlight_cases_list),
                    edgecolors=None, c='C1', alpha=0.9, **plotkwargs)

    if cov_plot:
        if cov_p is None:
            cov_p = conf_ellipsoid_sig2pct(3.0, df=2)  # 3-sigma for 2D gaussian
        cov_p_list = get_list(cov_p)
        for p in cov_p_list:
            plot_2d_cov_ellipse(ax=ax, varx=varx, vary=vary, p=p)

    if rug_plot:
        all_cases = set(cases_list) | set(highlight_cases_list)
        plot_rug_marks(ax, orientation=PlotOrientation.VERTICAL,
                       nums=slice_by_index(varx_points, all_cases))
        plot_rug_marks(ax, orientation=PlotOrientation.HORIZONTAL,
                       nums=slice_by_index(vary_points, all_cases))

    plt.xlabel(varx.name)
    plt.ylabel(vary.name)
    apply_category_labels(ax, varx, vary)
    plt.title(title)

    return fig, ax



def plot_2d_line(varx : InVar | OutVar,
                 vary : InVar | OutVar,
                 cases           : None | int | Iterable[int] = None,
                 highlight_cases : None | int | Iterable[int] = empty_list(),
                 invar_space : InVarSpace | Iterable[InVarSpace] = InVarSpace.NUMS,
                 ax     : Optional[Axes] = None,
                 title  : str            = '',
                 plotkwargs : dict       = dict(),
                 ) -> tuple[Figure, Axes]:
    """
    Plot an ensemble of 2D lines for two nonscalar variables.

    Parameters
    ----------
    varx : monaco.mc_var.InVar | monaco.mc_var.OutVar
        The x variable to plot.
    vary : monaco.mc_var.InVar | monaco.mc_var.OutVar
        The y variable to plot.
    cases : None | int | Iterable[int], default: None
        The cases to plot. If None, then all cases are highlighted.
    highlight_cases : None | int | Iterable[int], default: []
        The cases to highlight. If [], then no cases are highlighted.
    invar_space : monaco.InVarSpace | Iterable[InVarSpace], default: 'nums'
        The space to plot invars in, either 'nums' or 'pcts'. If an iterable,
        specifies this individually for each of varx, vary, and varz.
    ax : matplotlib.axes.Axes, default: None
        The axes handle to plot in. If None, a new figure is created.
    title : str, default: ''
        The figure title.

    Returns
    -------
    (fig, ax) : (matplotlib.figure.Figure, matplotlib.axes.Axes)
        fig is the figure handle for the plot.
        ax is the axes handle for the plot.
    """
    fig, ax = manage_axis(ax, is3d=False)
    invar_space = manage_invar_space(invar_space=invar_space, nvars=2)

    cases_list = get_cases(varx.ncases, cases)
    highlight_cases_list = get_cases(varx.ncases, highlight_cases)
    reg_cases = set(cases_list) - set(highlight_cases_list)
    varx_points = get_plot_points(varx, invar_space[0])
    vary_points = get_plot_points(vary, invar_space[1])
    for i in reg_cases:
        plt.plot(varx_points[i], vary_points[i],
                 linestyle='-', color='black', alpha=0.2, **plotkwargs)
    for i in highlight_cases_list:
        plt.plot(varx_points[i], vary_points[i],
                 linestyle='-', color='C1', alpha=0.9, **plotkwargs)

    for varstat in vary.varstats:
        varx_points_max = max(varx_points, key=len)
        if length(varstat.nums[0]) == 1:  # Single Statistic
            plt.plot(varx_points_max, varstat.nums[:],
                     linestyle='-', color='C0', alpha=0.9, **plotkwargs)
        elif length(varstat.nums[0]) == 3:  # Sided Order Statistic
            plt.plot(varx_points_max, varstat.nums[:, 1],
                     linestyle='-', color='C0', alpha=0.9, **plotkwargs)

        if varstat.confidence_interval_high_nums is not None:
            ax.fill_between(varx_points_max,
                            varstat.confidence_interval_low_nums,
                            varstat.confidence_interval_high_nums,
                            color='C0', alpha=0.3)
        if length(varstat.nums[0]) in (2, 3):  # Sided Order Statistic
            ax.fill_between(varx_points_max,
                            np.array(varstat.nums)[:, 0],
                            np.array(varstat.nums)[:, -1],
                            color='C0', alpha=0.3)

    plt.xlabel(varx.name)
    plt.ylabel(vary.name)
    apply_category_labels(ax, varx, vary)
    plt.title(title)

    return fig, ax



def plot_3d_scatter(varx : InVar | OutVar,
                    vary : InVar | OutVar,
                    varz : InVar | OutVar,
                    cases           : None | int | Iterable[int] = None,
                    highlight_cases : None | int | Iterable[int] = empty_list(),
                    invar_space : InVarSpace | Iterable[InVarSpace] = InVarSpace.NUMS,
                    ax     : Optional[Axes] = None,
                    title  : str            = '',
                    plotkwargs : dict       = dict(),
                    ) -> tuple[Figure, Axes]:
    """
    Plot a scatter plot of three variables in 3D space.

    Parameters
    ----------
    varx : monaco.mc_var.InVar | monaco.mc_var.OutVar
        The x variable to plot.
    vary : monaco.mc_var.InVar | monaco.mc_var.OutVar
        The y variable to plot.
    varz : monaco.mc_var.InVar | monaco.mc_var.OutVar
        The z variable to plot.
    cases : None | int | Iterable[int], default: None
        The cases to plot. If None, then all cases are highlighted.
    highlight_cases : None | int | Iterable[int], default: []
        The cases to highlight. If [], then no cases are highlighted.
    invar_space : monaco.InVarSpace | Iterable[InVarSpace], default: 'nums'
        The space to plot invars in, either 'nums' or 'pcts'. If an iterable,
        specifies this individually for each of varx, vary, and varz.
    ax : matplotlib.axes.Axes, default: None
        The axes handle to plot in. If None, a new figure is created.
    title : str, default: ''
        The figure title.

    Returns
    -------
    (fig, ax) : (matplotlib.figure.Figure, matplotlib.axes.Axes)
        fig is the figure handle for the plot.
        ax is the axes handle for the plot.
    """
    fig, ax = manage_axis(ax, is3d=True)
    invar_space = manage_invar_space(invar_space=invar_space, nvars=3)

    cases_list = get_cases(varx.ncases, cases)
    highlight_cases_list = get_cases(varx.ncases, highlight_cases)
    reg_cases = set(cases_list) - set(highlight_cases_list)
    varx_points = get_plot_points(varx, invar_space[0])
    vary_points = get_plot_points(vary, invar_space[1])
    varz_points = get_plot_points(varz, invar_space[2])
    if reg_cases:
        ax.scatter(slice_by_index(varx_points, reg_cases),
                   slice_by_index(vary_points, reg_cases),
                   slice_by_index(varz_points, reg_cases),
                   edgecolors=None, c='k', alpha=0.4, **plotkwargs)
    if highlight_cases_list:
        ax.scatter(slice_by_index(varx_points, highlight_cases_list),
                   slice_by_index(vary_points, highlight_cases_list),
                   slice_by_index(varz_points, highlight_cases_list),
                   edgecolors=None, c='C1', alpha=0.9, **plotkwargs)

    ax.set_xlabel(varx.name)
    ax.set_ylabel(vary.name)
    ax.set_zlabel(varz.name)
    apply_category_labels(ax, varx, vary, varz)
    plt.title(title)

    return fig, ax



def plot_2p5d_line(varx : InVar | OutVar,
                   vary : InVar | OutVar,
                   varz : InVar | OutVar,
                   cases           : None | int | Iterable[int] = None,
                   highlight_cases : None | int | Iterable[int] = empty_list(),
                   invar_space : InVarSpace | Iterable[InVarSpace] = InVarSpace.NUMS,
                   ax     : Optional[Axes] = None,
                   title  : str            = '',
                   plotkwargs : dict       = dict(),
                   ) -> tuple[Figure, Axes]:
    """
    Plot an ensemble of 2.5D lines for one scalar and two nonscalar variables.

    Parameters
    ----------
    varx : monaco.mc_var.InVar | monaco.mc_var.OutVar
        The x variable to plot.
    vary : monaco.mc_var.InVar | monaco.mc_var.OutVar
        The y variable to plot.
    varz : monaco.mc_var.InVar | monaco.mc_var.OutVar
        The z variable to plot.
    cases : None | int | Iterable[int], default: None
        The cases to plot. If None, then all cases are highlighted.
    highlight_cases : None | int | Iterable[int], default: []
        The cases to highlight. If [], then no cases are highlighted.
    invar_space : monaco.InVarSpace | Iterable[InVarSpace], default: 'nums'
        The space to plot invars in, either 'nums' or 'pcts'. If an iterable,
        specifies this individually for each of varx, vary, and varz.
    ax : matplotlib.axes.Axes, default: None
        The axes handle to plot in. If None, a new figure is created.
    title : str, default: ''
        The figure title.

    Returns
    -------
    (fig, ax) : (matplotlib.figure.Figure, matplotlib.axes.Axes)
        fig is the figure handle for the plot.
        ax is the axes handle for the plot.
    """
    var0 = deepcopy(varx)
    var1 = deepcopy(vary)
    var2 = deepcopy(varz)
    npoints = 0
    for var in (var0, var1, var2):
        if var.maxdim == 1:
            npoints = max(npoints, max(len(num) for num in var.nums))
    for var in (var0, var1, var2):
        if var.maxdim == 0:
            for i in range(var.ncases):
                var.nums[i] = np.array([var.nums[i] for _ in range(npoints)])
                if isinstance(var, monaco.mc_var.InVar):
                    var.pcts[i] = [float(var.pcts[i]) for _ in range(npoints)]
            var.maxdim = 1

    fig, ax = plot_3d_line(varx=var0, vary=var1, varz=var2,
                           cases=cases, highlight_cases=highlight_cases,
                           invar_space=invar_space,
                           ax=ax, title=title, plotkwargs=plotkwargs)

    return fig, ax



def plot_3d_line(varx : InVar | OutVar,
                 vary : InVar | OutVar,
                 varz : InVar | OutVar,
                 cases           : None | int | Iterable[int] = None,
                 highlight_cases : None | int | Iterable[int] = empty_list(),
                 invar_space : InVarSpace | Iterable[InVarSpace] = InVarSpace.NUMS,
                 ax     : Optional[Axes] = None,
                 title  : str            = '',
                 plotkwargs : dict       = dict(),
                 ) -> tuple[Figure, Axes]:
    """
    Plot an ensemble of 3D lines for three nonscalar variables.

    Parameters
    ----------
    varx : monaco.mc_var.InVar | monaco.mc_var.OutVar
        The x variable to plot.
    vary : monaco.mc_var.InVar | monaco.mc_var.OutVar
        The y variable to plot.
    varz : monaco.mc_var.InVar | monaco.mc_var.OutVar
        The z variable to plot.
    cases : None | int | Iterable[int], default: None
        The cases to plot. If None, then all cases are highlighted.
    highlight_cases : None | int | Iterable[int], default: []
        The cases to highlight. If [], then no cases are highlighted.
    invar_space : monaco.InVarSpace | Iterable[InVarSpace], default: 'nums'
        The space to plot invars in, either 'nums' or 'pcts'. If an iterable,
        specifies this individually for each of varx, vary, and varz.
    ax : matplotlib.axes.Axes, default: None
        The axes handle to plot in. If None, a new figure is created.
    title : str, default: ''
        The figure title.

    Returns
    -------
    (fig, ax) : (matplotlib.figure.Figure, matplotlib.axes.Axes)
        fig is the figure handle for the plot.
        ax is the axes handle for the plot.
    """
    fig, ax = manage_axis(ax, is3d=True)
    invar_space = manage_invar_space(invar_space=invar_space, nvars=3)

    cases_list = get_cases(varx.ncases, cases)
    highlight_cases_list = get_cases(varx.ncases, highlight_cases)
    reg_cases = set(cases_list) - set(highlight_cases_list)
    varx_points = get_plot_points(varx, invar_space[0])
    vary_points = get_plot_points(vary, invar_space[1])
    varz_points = get_plot_points(varz, invar_space[2])
    for i in reg_cases:
        ax.plot(varx_points[i], vary_points[i], varz_points[i],
                linestyle='-', color='black', alpha=0.3, **plotkwargs)
    for i in highlight_cases_list:
        ax.plot(varx_points[i], vary_points[i], varz_points[i],
                linestyle='-', color='C1', alpha=0.9, **plotkwargs)

    ax.set_xlabel(varx.name)
    ax.set_ylabel(vary.name)
    ax.set_zlabel(varz.name)
    apply_category_labels(ax, varx, vary, varz)
    plt.title(title)

    return fig, ax



def plot_cov_corr(matrix     : np.ndarray,
                  varnames   : list[str],
                  ax         : Optional[Axes] = None,
                  title      : str            = '',
                  plotkwargs : dict           = dict(),
                  ) -> tuple[Figure, Axes]:
    """
    Plot either a covariance or correlation matrix.

    Parameters
    ----------
    matrix : numpy.ndarray
        The covariance or correlation matrix.
    varnames : list[str]
        A list of the variable names.
    ax : matplotlib.axes.Axes, default: None
        The axes handle to plot in. If None, a new figure is created.
    title : str, default: ''
        The figure title.

    Returns
    -------
    (fig, ax) : (matplotlib.figure.Figure, matplotlib.axes.Axes)
        fig is the figure handle for the plot.
        ax is the axes handle for the plot.
    """
    fig, ax = manage_axis(ax, is3d=False)
    # for a correlation matrix scale will always be 1 from diagonal
    scale = np.nanmax(np.abs(matrix))
    im = ax.imshow(matrix, cmap="RdBu", vmin=-scale, vmax=scale, **plotkwargs)
    n = matrix.shape[1]

    ax.set_xticks(np.arange(n))
    ax.set_yticks(np.arange(n))
    ax.set_xticklabels(varnames)
    ax.set_yticklabels(varnames)
    ax.tick_params(top=True, bottom=False, labeltop=True, labelbottom=False)
    plt.setp(ax.get_xticklabels(), rotation=-30, ha='right', rotation_mode='anchor')

    for edge, spine in ax.spines.items():
        spine.set_visible(False)

    ax.set_xticks(np.arange(n+1)-.5, minor=True)
    ax.set_yticks(np.arange(n+1)-.5, minor=True)
    ax.grid(which='minor', color='w', linestyle='-', linewidth=3)
    ax.tick_params(which='minor', bottom=False, left=False)

    threshold = 0.6
    textcolors = ["black", "white"]
    kw = {'horizontalalignment': 'center', 'verticalalignment': 'center'}

    texts = []
    for i in range(n):
        for j in range(n):
            kw.update(color=textcolors[int(abs(matrix[i, j]/scale) > threshold)])
            text = im.axes.text(j, i, f'{matrix[i, j]:.2f}', **kw)
            texts.append(text)
    plt.title(title)

    return fig, ax



def plot_integration_convergence(outvar       : OutVar,
                                 dimension    : int,
                                 volume       : float,
                                 refval       : float          = None,
                                 conf         : float          = 0.95,
                                 samplemethod : SampleMethod   = SampleMethod.RANDOM,
                                 ax           : Optional[Axes] = None,
                                 title        : str            = '',
                                 plotkwargs   : dict           = dict(),
                                 ) -> tuple[Figure, Axes]:
    """
    For a Monte Carlo integration, plot the running integration estimate along
    with error bars for a given confidence level.

    Parameters
    ----------
    outvar : monaco.mc_var.OutVar
        The variable representing the integration estimate.
    dimension : int
        The number of dimensions over which the integration was performed.
    volume : float
        The total integration volume.
    refval : float, default: None
        If known a-priori, the reference value for the integration.
    conf : float, default: 0.95
        The confidence level for the error estimate. 0.95 corresponds to 95%.
    samplemethod : monaco.mc_enums.SampleMethod
        The sample method used for integration. Either 'random' or 'sobol'.
    ax : matplotlib.axes.Axes, default: None
        The axes handle to plot in. If None, a new figure is created.
    title : str, default: ''
        The figure title.

    Returns
    -------
    (fig, ax) : (matplotlib.figure.Figure, matplotlib.axes.Axes)
        fig is the figure handle for the plot.
        ax is the axes handle for the plot.
    """
    fig, ax = manage_axis(ax, is3d=False)

    if refval is not None:
        ax.axhline(refval, color='k')

    cummean = volume*np.cumsum(outvar.nums)/np.arange(1, outvar.ncases+1)
    err = integration_error(nums=np.array(outvar.nums), dimension=dimension, volume=volume,
                            conf=conf, samplemethod=samplemethod, runningerror=True)
    ax.plot(cummean, 'C1')
    ax.plot(cummean + err, 'C0')
    ax.plot(cummean - err, 'C0')

    ax.set_xlabel('Sample #')
    ax.set_ylabel(f'Convergence of {outvar.name} Integral')
    plt.title(title)

    return fig, ax



def plot_integration_error(outvar       : OutVar,
                           dimension    : int,
                           volume       : float,
                           refval       : float,
                           conf         : float          = 0.95,
                           samplemethod : SampleMethod   = SampleMethod.RANDOM,
                           ax           : Optional[Axes] = None,
                           title        : str            = '',
                           plotkwargs   : dict           = dict(),
                           ) -> tuple[Figure, Axes]:
    """
    For a Monte Carlo integration where the reference value is known, plot the
    running integration error along with error bounds for a given confidence
    level.

    Parameters
    ----------
    outvar : monaco.mc_var.OutVar
        The variable representing the integration estimate.
    dimension : int
        The number of dimensions over which the integration was performed.
    volume : float
        The total integration volume.
    refval : float
        The reference value for the integration.
    conf : float, default: 0.95
        The confidence level for the error estimate. 0.95 corresponds to 95%.
    samplemethod : monaco.mc_enums.SampleMethod
        The sample method used for integration. Either 'random' or 'sobol'.
    ax : matplotlib.axes.Axes, default: None
        The axes handle to plot in. If None, a new figure is created.
    title : str, default: ''
        The figure title.

    Returns
    -------
    (fig, ax) : (matplotlib.figure.Figure, matplotlib.axes.Axes)
        fig is the figure handle for the plot.
        ax is the axes handle for the plot.
    """
    fig, ax = manage_axis(ax, is3d=False)

    cummean = volume*np.cumsum(outvar.nums)/np.arange(1, outvar.ncases+1)
    err = integration_error(nums=np.array(outvar.nums), dimension=dimension, volume=volume,
                            conf=conf, samplemethod=samplemethod, runningerror=True)
    ax.loglog(err, 'C0')
    ax.plot(np.abs(cummean - refval), 'C1')

    ax.set_xlabel('Sample #')
    ax.set_ylabel(f'{outvar.name} {round(conf*100, 2)}% Confidence Error')
    plt.title(title)

    return fig, ax


def plot_sensitivities(outvar        : OutVar,
                       sensitivities : Sensitivities  = Sensitivities.RATIOS,
                       sort          : bool           = True,
                       ax            : Optional[Axes] = None,
                       title         : str            = '',
                       plotkwargs    : dict           = dict(),
                       ) -> tuple[Figure, Axes]:
    """
    For a Monte Carlo integration where the reference value is known, plot the
    running integration error along with error bounds for a given confidence
    level.

    Parameters
    ----------
    outvar : monaco.mc_var.OutVar
        The variable representing the integration estimate.
    sensitivities : monaco.mc_enums.Sensitivities, default: 'ratios'
        The sensitivities to plot. Either 'ratios' or 'indices'.
    sort : bool, default: True
        Whether to show the sensitivity indices sorted from greatest to least.
    ax : matplotlib.axes.Axes, default: None
        The axes handle to plot in. If None, a new figure is created.
    title : str, default: ''
        The figure title.

    Returns
    -------
    (fig, ax) : (matplotlib.figure.Figure, matplotlib.axes.Axes)
        fig is the figure handle for the plot.
        ax is the axes handle for the plot.
    """
    fig, ax = manage_axis(ax, is3d=False)
    fig.set_layout_engine('tight')

    if sensitivities == Sensitivities.RATIOS:
        sensitivities_dict = outvar.sensitivity_ratios
        ax.set_xlabel(f"Sensitivity Ratio for '{outvar.name}'")

    elif sensitivities == Sensitivities.INDICES:
        sensitivities_dict = outvar.sensitivity_indices
        ax.set_xlabel(f"Sensitivity Index for '{outvar.name}'")

    y_pos = np.arange(len(sensitivities_dict))
    if sort:
        sens_tuples = sorted(sensitivities_dict.items(), key=lambda item: item[1])
    else:
        sens_tuples = list(sensitivities_dict.items())
        y_pos = np.flipud(y_pos)
    invarnames = [invarname for invarname, _ in sens_tuples]
    sensitivities_vals = [val for _, val in sens_tuples]

    ax.barh(y_pos, sensitivities_vals, facecolor='k', alpha=0.5, tick_label=invarnames)
    plt.title(title)

    return fig, ax


def manage_axis(ax   : Optional[Axes],
                is3d : bool = False,
                ) -> tuple[Figure, Axes]:
    """
    Set the target axis, either by making a new figure or setting active an
    existing axis.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The target axis. If None, a new figure is created.
    is3d : bool, default: False
        If creating a new figure, whether the plot is a 3D plot.

    Returns
    -------
    (fig, ax) : (matplotlib.figure.Figure, matplotlib.axes.Axes)
        fig is the figure handle for the plot.
        ax is the axes handle for the plot.
    """
    if ax is not None:
        fig = ax.figure
    else:
        if is3d:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')
        else:
            fig, ax = plt.subplots(1, 1)
    plt.sca(ax)
    return fig, ax



def apply_category_labels(ax   : Axes,
                          varx : InVar | OutVar | None = None,
                          vary : InVar | OutVar | None = None,
                          varz : InVar | OutVar | None = None,
                          ) -> None:
    """
    For nonnumeric Monte Carlo variables, use the `nummap` to label the axes.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The target axis.
    varx : monaco.mc_var.InVar | monaco.mc_var.OutVar, default: None
        The x variable.
    vary : monaco.mc_var.InVar | monaco.mc_var.OutVar, default: None
        The y variable.
    varz : monaco.mc_var.InVar | monaco.mc_var.OutVar, default: None
        The z variable.
    """
    # Wrapped in try statements in case some categories aren't printable
    if varx is not None and varx.nummap is not None:
        try:
            ax.set_xticks(list(varx.nummap.keys()))
            ax.set_xticklabels(list(varx.nummap.values()))
        except Exception:
            pass
    if vary is not None and vary.nummap is not None:
        try:
            ax.set_yticks(list(vary.nummap.keys()))
            ax.set_yticklabels(list(vary.nummap.values()))
        except Exception:
            pass
    if varz is not None and varz.nummap is not None:
        try:
            ax.set_zticks(list(varz.nummap.keys()))
            ax.set_zticklabels(list(varz.nummap.values()))
        except Exception:
            pass


def get_hist_lim(ax          : Axes,
                 orientation : PlotOrientation,
                 ) -> tuple[float, float]:
    """
    Get the axis limits for a histogram.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The target axis.
    orientation : PlotOrientation
        The orientation of the histogram plot, either 'vertical' or
        'horizontal'.

    Returns
    -------
    lim : (float, float)
        Returns the (low, high) limits of the axis.
    """
    if orientation == PlotOrientation.VERTICAL:
        lim = ax.get_xlim()
    elif orientation == PlotOrientation.HORIZONTAL:
        lim = ax.get_ylim()
    return lim



def plot_rug_marks(ax          : Axes,
                   orientation : PlotOrientation,
                   nums        : Iterable[float]
                   ) -> None:
    """
    Plot rug marks for a histogram or scatter plot.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The target axis.
    orientation : PlotOrientation
        The orientation of the plot, either 'vertical' or 'horizontal'.
    nums : Iterable[float]
        The numbers to plot the rug marks at.
    """
    if ax is None:
        return

    xlim = ax.get_xlim()
    ylim = ax.get_ylim()
    if orientation == PlotOrientation.VERTICAL:
        for num in nums:
            plt.plot([num, num], [ylim[0], ylim[0] + 0.02*(ylim[1] - ylim[0])],
                     color='black', linewidth=0.5, alpha=0.5)
    elif orientation == PlotOrientation.HORIZONTAL:
        for num in nums:
            plt.plot([xlim[0], xlim[0] + 0.02*(xlim[1] - xlim[0])], [num, num],
                     color='black', linewidth=0.5, alpha=0.5)



def plot_2d_cov_ellipse(ax   : Axes,
                        varx : InVar | OutVar,
                        vary : InVar | OutVar,
                        p    : float,
                        ) -> None:
    """
    Add a covariance ellipse to a 2D scatter plot.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The target axis.
    varx : monaco.mc_var.InVar | monaco.mc_var.OutVar
        The x variable.
    vary : monaco.mc_var.InVar | monaco.mc_var.OutVar
        The y variable.
    p : float
        Coviariance percentile, assuming a gaussian distribution.
    """
    if ax is None:
        return

    # See https://www.visiondummy.com/2014/04/draw-error-ellipse-representing-covariance-matrix/
    allnums = [np.array(varx.nums), np.array(vary.nums)]
    center = [np.mean(varx.nums), np.mean(vary.nums)]

    covs = np.cov(np.array(allnums))
    eigvals, eigvecs = np.linalg.eigh(covs)  # Use eigh over eig since covs is guaranteed symmetric
    inds = (-eigvals).argsort()  # sort from largest to smallest
    eigvals = eigvals[inds]
    eigvecs = eigvecs[inds]
    angle = np.arctan2(eigvecs[0][1], eigvecs[0][0])*180/np.pi

    scalefactor = chi2.ppf(p, df=2)
    ellipse_axis_radii = np.sqrt(scalefactor*eigvals)

    ellipse = Ellipse(center, 2*ellipse_axis_radii[0], 2*ellipse_axis_radii[1], angle=angle,
                      fill=False, edgecolor='k')
    ax.add_patch(ellipse)

    # For now plot both eigenaxes
    plt.axline(xy1=(center[0], center[1]), slope=eigvecs[0][1]/eigvecs[0][0], color='k')
    plt.axline(xy1=(center[0], center[1]), slope=eigvecs[1][1]/eigvecs[1][0], color='k')



def get_cases(ncases : int,
              cases  : None | int | Iterable[int],
              ) -> list[int]:
    """
    Parse the `cases` input for plotting functions. If None, return a list of
    all the cases. Otherwise, return a list of all the specified cases.

    Parameters
    ----------
    ncases : int
        The total number of cases.
    cases : None | int | Iterable[int]
        The cases to downselect to.

    Returns
    -------
    cases_list : list[int]
        The cases.
    """
    if cases is None:
        cases = list(range(ncases))
    cases_list = get_list(cases)
    return cases_list



def manage_invar_space(invar_space : InVarSpace | Iterable[InVarSpace],
                       nvars       : int,
                       ) -> list[InVarSpace]:
    '''
    Parse the `invarspace` input for plotting functions. If already an iterable
    of the right length, passes through. If a single value, returns a list of
    the invarspace of the right length.

    Parameters
    ----------
    invar_space : monaco.InVarSpace | Iterable[InVarSpace]
        The space to plot invars in, either 'nums' or 'pcts'. If an iterable,
        specifies this individually for each of varx, vary, and varz.
    nvars : int
        The length of the list to generate.

    Returns
    -------
    invar_space_list : list[InVarSpace]
        A list of length `nvars` for the desired InVarSpace's.
    '''
    if isinstance(invar_space, InVarSpace):
        invar_space = [invar_space for _ in range(nvars)]

    if len(invar_space) != nvars:
        raise ValueError(f'Argument invar_space = {invar_space} expected length {nvars}')
    invar_space_list = list(invar_space)

    return invar_space_list



def get_plot_points(var         : InVar | OutVar,
                    invar_space : InVarSpace,
                    ) -> list[float]:
    '''
    Get the points to plot based on the invar_space. Returns var.nums, unless
    var is an InVar and invar_space is 'pcts' in which case it returns
    var.pcts.

    Parameters
    ----------
    nvars : InVar | OutVar
        The target variable.
    invar_space : monaco.InVarSpace
        The space to plot the invars in.

    Returns
    -------
    plot_points : list[float]
        The points to plot.
    '''
    plot_points = var.nums
    if isinstance(var, monaco.mc_var.InVar) and invar_space == InVarSpace.PCTS:
        plot_points = var.pcts
    return plot_points



def get_var_steps(var : InVar | OutVar) -> OutVar:
    '''
    For a 1D variable, get an OutVar that has as values the simulation sets.

    Parameters
    ----------
    var : InVar | OutVar
        The target variable.

    Returns
    -------
    varsteps : OutVar
        The variable with simulation steps as values.
    '''
    vals = [np.arange(len(num)) for num in var.nums]
    varsteps = monaco.mc_var.OutVar(name='Simulation Steps', vals=vals)
    return varsteps



# mc_multi_plot.py
from __future__ import annotations

import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.axes import Axes
from monaco.mc_plot import plot_hist, plot_2d_scatter
from monaco.mc_var import InVar, OutVar
from monaco.mc_enums import PlotOrientation, InVarSpace
from monaco.helper_functions import empty_list, get_list
from typing import Optional, Iterable


def multi_plot(vars        : list[InVar | OutVar],
               cases           : None | int | Iterable[int] = None,
               highlight_cases : None | int | Iterable[int] = empty_list(),
               rug_plot    : bool   = False,
               cov_plot    : bool   = False,
               cov_p       : None | float | Iterable[float] = None,
               invar_space : InVarSpace | Iterable[InVarSpace] = InVarSpace.NUMS,
               fig         : Figure | None = None,
               title       : str    = '',
               plotkwargs  : dict   = dict(),
               ) -> tuple[Figure, tuple[Axes, ...]]:
    """
    Umbrella function to make more complex plots of Monte Carlo variables.

    Parameters
    ----------
    vars : list[monaco.mc_var.InVar | monaco.mc_var.OutVar]
        The variables to plot.
    cases : None | int | Iterable[int], default: None
        The cases to plot. If None, then all cases are plotted.
    highlight_cases : None | int | Iterable[int], default: []
        The cases to highlight. If [], then no cases are highlighted.
    rug_plot : bool, default: False
        Whether to plot rug marks.
    cov_plot : bool, default: False
        Whether to plot a covariance ellipse at a certain gaussian percentile
        level.
    cov_p : None | float | Iterable[float], default: None
        The gaussian percentiles for the covariance plot.
    invar_space : monaco.InVarSpace | Iterable[InVarSpace], default: 'nums'
        The space to plot invars in, either 'nums' or 'pcts'. If an iterable,
        specifies this individually for each of varx, vary, and varz.
    fig : matplotlib.figure.Figure, default: None
        The figure handle to plot in. If None, a new figure is created.
    title : str, default: ''
        The figure title.

    Returns
    -------
    (fig, axes) : (matplotlib.figure.Figure, (matplotlib.axes.Axes, ...))
        fig is the figure handle for the plot.
        axes is a tuple of the axes handles for the plots.
    """
    # Split larger vars
    if len(vars) == 1:
        if isinstance(vars[0], OutVar) and vars[0].maxdim == 1:
            var_split = vars[0].split()
            origname = vars[0].name
            varx = var_split[origname + ' [0]']
            vary = var_split[origname + ' [1]']
            vars = [varx, vary]
        else:
            raise ValueError(f'Invalid vars[0] dimension: {vars[0].maxdim}')

    # Two Variable Plots
    if len(vars) == 2:
        if vars[0].isscalar and vars[1].isscalar:
            fig, axs = multi_plot_2d_scatter_hist(varx=vars[0], vary=vars[1],
                                                  cases=cases, highlight_cases=highlight_cases,
                                                  rug_plot=rug_plot,
                                                  cov_plot=cov_plot, cov_p=cov_p,
                                                  cumulative=False,
                                                  invar_space=invar_space,
                                                  fig=fig, title=title, plotkwargs=plotkwargs)
        else:
            raise ValueError( 'Invalid variable dimensions: ' +
                             f'{varx.name} {varx.maxdim}, ' +
                             f'{vary.name} {vary.maxdim}')

    # Many Variable Plots
    elif len(vars) > 2:
        scalarvars = [var for var in vars if var.isscalar]
        fig, axs = multi_plot_grid_tri(vars=scalarvars,
                                       cases=cases, highlight_cases=highlight_cases,
                                       rug_plot=rug_plot,
                                       cov_plot=cov_plot, cov_p=cov_p,
                                       cumulative=False,
                                       invar_space=invar_space,
                                       fig=fig, title=title, plotkwargs=plotkwargs)

    return fig, axs



def multi_plot_2d_scatter_hist(varx     : InVar | OutVar,
                               vary     : InVar | OutVar,
                               cases           : None | int | Iterable[int] = None,
                               highlight_cases : None | int | Iterable[int] = empty_list(),
                               rug_plot   : bool   = False,
                               cov_plot   : bool   = True,
                               cov_p      : None | float | Iterable[float] = None,
                               cumulative : bool   = False,
                               invar_space : InVarSpace | Iterable[InVarSpace] = InVarSpace.NUMS,
                               fig        : Figure | None = None,
                               title      : str    = '',
                               plotkwargs : dict   = dict(),
                               ) -> tuple[Figure, tuple[Axes, ...]]:
    """
    Plot two variables against each other with a central scatterplot and two
    histograms along the x and y axes.

    Parameters
    ----------
    varx : monaco.mc_var.InVar | monaco.mc_var.OutVar
        The x variable to plot.
    vary : monaco.mc_var.InVar | monaco.mc_var.OutVar
        The y variable to plot.
    cases : None | int | Iterable[int], default: None
        The cases to plot. If None, then all cases are plotted.
    highlight_cases : None | int | Iterable[int], default: []
        The cases to highlight. If [], then no cases are highlighted.
    rug_plot : bool, default: False
        Whether to plot rug marks.
    cov_plot : bool, default: False
        Whether to plot a covariance ellipse at a certain gaussian percentile
        level.
    cov_p : None | float | Iterable[float], default: None
        The gaussian percentiles for the covariance plot.
    cumulative : bool, default: False
        Whether to plot the histograms as cumulative distribution functions.
    invar_space : monaco.InVarSpace | Iterable[InVarSpace], default: 'nums'
        The space to plot invars in, either 'nums' or 'pcts'. If an iterable,
        specifies this individually for each of varx, vary, and varz.
    fig : matplotlib.figure.Figure, default: None
        The figure handle to plot in. If None, a new figure is created.
    title : str, default: ''
        The figure title.

    Returns
    -------
    (fig, (ax1, ax2, ax3)) : (matplotlib.figure.Figure,
    (matplotlib.axes.Axes, matplotlib.axes.Axes, matplotlib.axes.Axes))
        fig is the figure handle for the plot.
        (ax1, ax2, ax3) are the axes handles for the central, y-axis, and
        x-axis plots, respectively.
    """
    fig = handle_fig(fig)

    gs = fig.add_gridspec(4, 4)
    ax1 = fig.add_subplot(gs[3, 1:4])
    ax2 = fig.add_subplot(gs[0:3, 0])
    ax3 = fig.add_subplot(gs[0:3, 1:4], sharex=ax1, sharey=ax2)

    plot_hist(varx, cases=cases, highlight_cases=highlight_cases,
              cumulative=cumulative, rug_plot=False, invar_space=invar_space,
              ax=ax1, title='', plotkwargs=plotkwargs)
    plot_hist(vary, cases=cases, highlight_cases=highlight_cases,
              cumulative=cumulative, rug_plot=False, invar_space=invar_space,
              ax=ax2, title='', plotkwargs=plotkwargs,
              orientation=PlotOrientation.HORIZONTAL)
    plot_2d_scatter(varx, vary,
                    cases=cases, highlight_cases=highlight_cases,
                    rug_plot=rug_plot,
                    cov_plot=cov_plot, cov_p=cov_p,
                    invar_space=invar_space,
                    ax=ax3, title='', plotkwargs=plotkwargs)

    ax1.set_ylabel('')
    ax2.set_xlabel('')
    ax3.set_xlabel('')
    ax3.set_ylabel('')
    ax3.xaxis.set_tick_params(labelbottom=False)
    ax3.yaxis.set_tick_params(labelbottom=False)

    plt.suptitle(title, y=.93)

    return fig, (ax1, ax2, ax3)



def multi_plot_grid_tri(vars     : list[InVar | OutVar],
                        cases           : None | int | Iterable[int] = None,
                        highlight_cases : None | int | Iterable[int] = empty_list(),
                        rug_plot   : bool   = False,
                        cov_plot   : bool   = False,
                        cov_p      : None | float | Iterable[float] = None,
                        cumulative : bool   = False,
                        invar_space : InVarSpace | Iterable[InVarSpace] = InVarSpace.NUMS,
                        fig        : Figure | None = None,
                        title      : str    = '',
                        plotkwargs : dict   = dict(),
                        ) -> tuple[Figure, tuple[Axes, ...]]:
    """
    Plot multiple variables against each other in a triangular grid. The
    off-diagonal grid locations show scatterplots of the two corresponding
    variables. The plots along the diagonal show histograms for the
    corresponding variables.

    Parameters
    ----------
    vars : list[monaco.mc_var.InVar | monaco.mc_var.OutVar]
        The variables to plot.
    cases : None | int | Iterable[int], default: None
        The cases to plot. If None, then all cases are plotted.
    highlight_cases : None | int | Iterable[int], default: []
        The cases to highlight. If [], then no cases are highlighted.
    rug_plot : bool, default: False
        Whether to plot rug marks.
    cov_plot : bool, default: False
        Whether to plot a covariance ellipse at a certain gaussian percentile
        level.
    cov_p : None | float | Iterable[float], default: None
        The gaussian percentiles for the covariance plot.
    cumulative : bool, default: False
        Whether to plot the histograms as cumulative distribution functions.
    invar_space : monaco.InVarSpace | Iterable[InVarSpace], default: 'nums'
        The space to plot invars in, either 'nums' or 'pcts'. If an iterable,
        specifies this individually for each of varx, vary, and varz.
    fig : matplotlib.figure.Figure, default: None
        The figure handle to plot in. If None, a new figure is created.
    title : str, default: ''
        The figure title.

    Returns
    -------
        (fig, axes) : (matplotlib.figure.Figure, (matplotlib.axes.Axes, ...))
            fig is the figure handle for the plot.
            axes is a tuple of the axes handles for the plots, starting from
            the top-left corner and working left-to-right, then top-to-bottom.
    """
    fig = handle_fig(fig)

    nvars = len(vars)
    axs = fig.subplots(nvars, nvars, sharex='col')
    for i in range(nvars):
        row_scatter_axs = [ax for k, ax in enumerate(axs[i]) if k < i]
        for j in range(nvars):
            ax = axs[i][j]
            if i == j:
                plot_hist(vars[i], cases=cases, highlight_cases=highlight_cases,
                          cumulative=cumulative, rug_plot=False, invar_space=invar_space,
                          ax=ax, title='', plotkwargs=plotkwargs)
                ax.yaxis.tick_right()
            elif j > i:
                fig.delaxes(ax)
            else:
                # Don't link the histogram y axis
                ax.get_shared_y_axes()._grouper.join(*row_scatter_axs)
                plot_2d_scatter(vars[j], vars[i],
                                cases=cases, highlight_cases=highlight_cases,
                                rug_plot=rug_plot, cov_plot=cov_plot, cov_p=cov_p,
                                invar_space=invar_space,
                                ax=ax, title='', plotkwargs=plotkwargs)
                if j > 0:
                    ax.set_yticklabels([])

            ax.set_xlabel('')
            ax.set_ylabel('')
            if j == 0:
                ax.set_ylabel(vars[i].name)
            if i == nvars-1:
                ax.set_xlabel(vars[j].name)

    fig.subplots_adjust(wspace=0.075, hspace=0.075)
    fig.suptitle(title)

    return fig, axs


def multi_plot_grid_rect(varsx     : InVar | OutVar | list[InVar | OutVar],
                         varsy     : InVar | OutVar | list[InVar | OutVar] = None,
                         cases           : None | int | Iterable[int] = None,
                         highlight_cases : None | int | Iterable[int] = empty_list(),
                         rug_plot   : bool   = False,
                         cov_plot   : bool   = False,
                         cov_p      : None | float | Iterable[float] = None,
                         cumulative : bool   = False,
                         invar_space : InVarSpace | Iterable[InVarSpace] = InVarSpace.NUMS,
                         fig        : Figure | None = None,
                         title      : str    = '',
                         plotkwargs : dict   = dict(),
                         ) -> tuple[Figure, tuple[Axes, ...]]:
    """
    Plot multiple variables against each other in a rectangual grid. The
    left column and bottom row show histograms for each variable, and the middle
    grid locations show the scatter plots between each pari of variables.

    Parameters
    ----------
    varsx : monaco.mc_var.InVar | monaco.mc_var.OutVar | list[InVar | OutVar]
        The variables to plot along the X axis.
    varsy : monaco.mc_var.InVar | monaco.mc_var.OutVar | list[InVar | OutVar]
        The variables to plot along the Y axis.
    cases : None | int | Iterable[int], default: None
        The cases to plot. If None, then all cases are plotted.
    highlight_cases : None | int | Iterable[int], default: []
        The cases to highlight. If [], then no cases are highlighted.
    rug_plot : bool, default: False
        Whether to plot rug marks.
    cov_plot : bool, default: False
        Whether to plot a covariance ellipse at a certain gaussian percentile
        level.
    cov_p : None | float | Iterable[float], default: None
        The gaussian percentiles for the covariance plot.
    cumulative : bool, default: False
        Whether to plot the histograms as cumulative distribution functions.
    invar_space : monaco.InVarSpace | Iterable[InVarSpace], default: 'nums'
        The space to plot invars in, either 'nums' or 'pcts'. If an iterable,
        specifies this individually for each of varx, vary, and varz.
    fig : matplotlib.figure.Figure, default: None
        The figure handle to plot in. If None, a new figure is created.
    title : str, default: ''
        The figure title.

    Returns
    -------
        (fig, axes) : (matplotlib.figure.Figure, (matplotlib.axes.Axes, ...))
            fig is the figure handle for the plot.
            axes is a tuple of the axes handles for the plots, starting from
            the top-left corner and working left-to-right, then top-to-bottom.
    """
    fig = handle_fig(fig)

    varsx = get_list(varsx)
    varsy = get_list(varsy)
    nvarsx = len(varsx)
    nvarsy = len(varsy)

    axs = fig.subplots(nvarsy+1, nvarsx+1, sharex='col')
    for i in range(nvarsy+1):
        row_scatter_axs = [ax for k, ax in enumerate(axs[i]) if k < nvarsy+1]
        for j in range(nvarsx+1):
            ax = axs[i][j]
            if i == nvarsy and j == 0:
                fig.delaxes(ax)
            elif i == nvarsy:
                plot_hist(varsx[j-1], cases=cases, highlight_cases=highlight_cases,
                          cumulative=cumulative, rug_plot=False, invar_space=invar_space,
                          orientation=PlotOrientation.VERTICAL,
                          ax=ax, title='', plotkwargs=plotkwargs)
            elif j == 0:
                plot_hist(varsy[i], cases=cases, highlight_cases=highlight_cases,
                          cumulative=cumulative, rug_plot=False, invar_space=invar_space,
                          orientation=PlotOrientation.HORIZONTAL,
                          ax=ax, title='', plotkwargs=plotkwargs)
                ax.set_xticklabels([])
            else:
                # Don't link the histogram y axis
                ax.get_shared_y_axes()._grouper.join(*row_scatter_axs)
                plot_2d_scatter(varsx[j-1], varsy[i],
                                cases=cases, highlight_cases=highlight_cases,
                                rug_plot=rug_plot, cov_plot=cov_plot, cov_p=cov_p,
                                invar_space=invar_space,
                                ax=ax, title='', plotkwargs=plotkwargs)
            if j > 0:
                ax.set_yticklabels([])

            ax.set_xlabel('')
            ax.set_ylabel('')
            if i < nvarsy and j == 0:
                ax.set_ylabel(varsy[i].name)
            if i == nvarsy and j != 0:
                ax.set_xlabel(varsx[j-1].name)

    fig.subplots_adjust(wspace=0.075, hspace=0.075)
    fig.suptitle(title)

    return fig, axs



def handle_fig(fig : Optional[Figure]) -> Figure:
    """
    Set the target figure, either by making a new figure or setting active an
    existing one.

    Parameters
    ----------
    fig : matplotlib.figure.Figure
        The target figure. If None, a new figure is created.

    Returns
    -------
    fig : matplotlib.figure.Figure
        The active figure.
    """
    if fig is None:
        fig = plt.figure()
    else:
        plt.figure(fig.num)
    plt.clf()
    return fig



# helper_functions.py
from __future__ import annotations

from operator import itemgetter
from tqdm import tqdm
import numpy as np
from typing import Callable, Any, Sequence, Iterable, Sized
from time import time
from functools import wraps
from hashlib import sha512
import warnings

# Optional imports
try:
    import pandas as pd
except ImportError:
    pd = None


def next_power_of_2(x : int) -> int:
    """
    Returns the next power of two greater than or equal to the input.

    Parameters
    ----------
    x : int
        The input number.

    Returns
    -------
    nextpow2 : int
        The next power of two, nextpow2 >= x.
    """
    if x <= 0:
        return 0
    else:
        return int(2**np.ceil(np.log2(x)))


def hash_str_repeatable(s : str) -> int:
    """
    By default, string hashing in python is randomized. This function returns a
    repeatable non-randomized hash for strings.
    See: https://docs.python.org/3/using/cmdline.html#cmdoption-R

    Parameters
    ----------
    s : str
        The string to hash.

    Returns
    -------
    s_hash : int
        The hash of str.
    """
    return int(sha512(s.encode('utf-8')).hexdigest(), 16)


def hashable_val(val : Any) -> Any:
    """
    For `nummap` and `valmap`, we need to use values as keys in a dictionary.
    This function will return the string representation of a value if that
    value is not hashable.

    Parameters
    ----------
    val : Any
        The value to hash.

    Returns
    -------
    hashable_val : Any
        A hashable representation of the val.
    """
    try:
        hash(val)
        return val
    except TypeError:
        return str(val)


def is_num(val : Any) -> bool:
    """
    Type checking function to see if the input is a number.

    Parameters
    ----------
    val : Any
        The value to check.

    Returns
    -------
    isnum : bool
        Returns True if the input is a number, False otherwise.
    """
    if isinstance(val, bool) or isinstance(val, str):
        return False
    else:
        try:
            float(val)
        except (ValueError, TypeError):
            return False
        else:
            return True


def length(x : Any) -> int | None:
    """
    Genericized length function that works on scalars (which have length 1).

    Parameters
    ----------
    x : Any
        The value to check.

    Returns
    -------
    x_len : int
        The length of the input. If not a sequence or scalar, returns None.
    """
    if isinstance(x, Sized):
        return len(x)
    elif isinstance(x, (np.float64, float, bool, int)):
        return 1
    else:
        return None


def get_list(x : Any) -> list[Any]:
    """
    Converts the input to an iterable list.

    Parameters
    ----------
    x : Any
        The object to convert.

    Returns
    -------
    x_list : list
        A list conversion of the input.
    """
    if x is None:
        return list()
    elif isinstance(x, str):
        return [x, ]
    elif pd and isinstance(x, pd.DataFrame):
        return [x, ]
    elif isinstance(x, Iterable):
        if isinstance(x, np.ndarray) and np.ndim(x) == 0:
            return [x[()], ]
        return list(x)
    else:
        return [x, ]


def slice_by_index(sequence : Sequence[Any],
                   indices  : int | Iterable[int],
                   ) -> list:
    """
    Returns a slice of a sequence at the specified indices.

    Parameters
    ----------
    sequence : Sequence
        The sequence to slice.
    indices : int | Iterable
        The indices to slice at.

    Returns
    -------
    slice : list
        A list representing the values of the input sequence at the specified
        indices.
    """
    indices_list = get_list(indices)
    if sequence is None or indices_list == list():
        return []
    items = itemgetter(*indices_list)(sequence)
    if len(indices_list) == 1:
        return [items]
    return list(items)


def vprint(verbose : bool, *args, **kwargs) -> None:
    """
    Print only if verbose is True.

    Parameters
    ----------
    verbose : bool
        Flag to determine whether to print something.
    *args, **kwargs
        Must include something to print here!
    """
    if verbose:
        print(*args, **kwargs)


def warn_short_format(message, category, filename, lineno, file=None, line=None) -> str:
    """
    Custom warning format for use in vwarn()
    """
    return f'{category.__name__}: {message}\n'


def vwarn(verbose : bool, *args, **kwargs) -> None:
    """
    Warn only if verbose is True.

    Parameters
    ----------
    verbose : bool
        Flag to determine whether to print something.
    *args, **kwargs
        Must include a warning message here!
    """
    if verbose:
        warn_default_format = warnings.formatwarning
        warnings.formatwarning = warn_short_format  # type: ignore
        warnings.warn(*args, **kwargs)
        warnings.formatwarning = warn_default_format


def vwrite(verbose : bool, *args, **kwargs) -> None:
    """
    Perform a tqdm.write() only if verbose is True.

    Parameters
    ----------
    verbose : bool
        Flag to determine whether to write something.
    *args, **kwargs
        Must include something to write here!
    """
    if verbose:
        tqdm.write(*args, **kwargs)


def timeit(fcn : Callable):
    """
    Function decorator to print out the function runtime in milliseconds.

    Parameters
    ----------
    fcn : Callable
        Function to time.
    """
    @wraps(fcn)
    def timed(*args, **kw):
        t0 = time()
        output = fcn(*args, **kw)
        t1 = time()
        print(f'"{fcn.__name__}" took {(t1 - t0)*1000 : .3f} ms to execute.\n')
        return output
    return timed


def empty_list() -> list:
    """
    Sentinel for default arguments being an empty list.

    Returns
    -------
    empty_list : list
        An empty list.
    """
    return []


def flatten(nested_x : Iterable[Any]) -> list[Any]:
    """
    Flattens a nested interable into a list with all nested items.

    Parameters
    ----------
    nested_x : Iterable
        Nested iterable.

    Returns
    -------
    flattened_x : list
        The nested iterable flattened into a list.
    """
    def flatten_generator(x):
        for element in x:
            if isinstance(element, Iterable) and not isinstance(element, (str, bytes)):
                yield from flatten(element)
            else:
                yield element

    flattened_x = list(flatten_generator(nested_x))
    return flattened_x
